unit BayesU;

//(*
//  BayesNet - Bayes Network Implementation
//
//  Sebastian Batton (s.batton@gmx.de)
//  Matthias Hupka (matthias-hupka@gmx.de)
//  Kevin Niehage (kevinniehage@gmx.de)
//  Johannes Weydt (johannes.weydt@gmx.de)
//
//  Duale Hochschule Mannheim - TAI06-ABC
//  Knowledge Based Systems - P.H. Vossen
//*)

interface

type
//>> DATA FILE
  TStringArray = array of String;

  TValueSet = record
    Name   : String;
    Values : TStringArray;
  end;
  TValueSets = array of TValueSet;

  TValueAssign = record
    Name  : String;
    Value : String;
  end;
  TValueAssigns = array of TValueAssign;

  TValueList = record
    Values : TValueAssigns;
    Result : String;
  end;
  TValueLists = array of TValueList;

  TValueCollection = record
    ValueSets  : TValueSets;
    ResultSet  : TStringArray;
    ValueLists : TValueLists;
  end;
//<< DATA FILE

//>> KNOWLEDGE BASE
  TResult = record
    Count : LongWord;
    Name  : String;
  end;
  TResults = array of TResult;

  TResultList = record
    Count  : LongWord;
    Result : String;
    Values : TResults;
  end;
  TResultLists = array of TResultList;

  TValue = record
    Name    : String;
    Results : TResults;
  end;
  TValues = array of TValue;

  TEntry = record
    Name   : String;
    Values : TValues;
  end;
  TEntries = array of TEntry;

  TKnowledgeBase = record
    Entries : TEntries;
    Results : TResults;
  end;
//<< KNOWLEDGE BASE

function CheckCollection(const ACollection : TValueCollection) : Boolean;
function GenerateKnowledgeBase(const ACollection : TValueCollection) : TKnowledgeBase;
function LoadCollection(const AFileName : String; out ACollection : TValueCollection) : Boolean;
function SelectCaseResults(const ABase : TKnowledgeBase; const ACase : TValueAssigns; out AResults : TResultLists) : Boolean;

implementation

uses
  SysUtils;

function CheckCollection(const ACollection : TValueCollection) : Boolean;
  function IsInSet(const AEntry : String; const ASet : TStringArray) : Boolean;
  var
    LIndex : LongInt;
  begin
    Result := false;
    for LIndex := 0 to Pred(Length(ASet)) do
    begin
      Result := AnsiLowerCase(AEntry) = AnsiLowerCase(ASet[LIndex]);
      if Result then
        Break;
    end;
  end;
var
  LIndexA : LongInt;
  LIndexB : LongInt;
  LIndexC : LongInt;
begin
  // minimal size of knowledge base
  Result := (Length(ACollection.ResultSet) > 1) and
            (Length(ACollection.ValueSets) > 0) and
            (Length(ACollection.ValueLists) > 0);
  if Result then
  begin
    // minimal size of knowledge base
    for LIndexA := 0 to Pred(Length(ACollection.ValueSets)) do
    begin
      Result := (Length(ACollection.ValueSets[LIndexA].Name) > 0) and
                (Length(ACollection.ValueSets[LIndexA].Values) > 1);
      if not(Result) then
      begin
        // ERROR
        Break;
      end;
    end;

    if Result then
    begin
      for LIndexA := 0 to Pred(Length(ACollection.ValueLists)) do
      begin
        // results of lists are in result set
        Result := false;
        for LIndexB := 0 to Pred(Length(ACollection.ResultSet)) do
        begin
          Result := IsInSet(ACollection.ValueLists[LIndexA].Result, ACollection.ResultSet);
          if Result then
            Break;
        end;

        if Result then
        begin
          for LIndexB := 0 to Pred(Length(ACollection.ValueSets)) do
          begin
            // values sets are found in all value list entries
            Result := false;
            for LIndexC := 0 to Pred(Length(ACollection.ValueLists[LIndexA].Values)) do
            begin
              Result := (AnsiLowerCase(ACollection.ValueSets[LIndexB].Name) = AnsiLowerCase(ACollection.ValueLists[LIndexA].Values[LIndexC].Name));

              if Result then
              begin
                // value of value list entry is in value set
                Result := IsInSet(ACollection.ValueLists[LIndexA].Values[LIndexC].Value, ACollection.ValueSets[LIndexB].Values);
                Break;
              end;
            end;

            if not(Result) then
            begin
              // ERROR
              Break;
            end;
          end;

          if not(Result) then
          begin
            // ERROR
            Break;
          end;
        end
        else
        begin
          // ERROR
          Break;
        end;
      end;
    end;
  end;
end;

function GenerateKnowledgeBase(const ACollection : TValueCollection) : TKnowledgeBase;
var
  LIndexA : LongInt;
  LIndexB : LongInt;
  LIndexC : LongInt;
  LIndexD : LongInt;
  LIndexE : LongInt;
begin
  // initialize results
  SetLength(Result.Results, Length(ACollection.ResultSet));
  for LIndexA := 0 to Pred(Length(ACollection.ResultSet)) do
  begin
    Result.Results[LIndexA].Count := 0;
    Result.Results[LIndexA].Name  := ACollection.ResultSet[LIndexA];

    for LIndexB := 0 to Pred(Length(ACollection.ValueLists)) do
    begin
      if (AnsiLowerCase(ACollection.ValueLists[LIndexB].Result) = AnsiLowerCase(Result.Results[LIndexA].Name)) then
        Inc(Result.Results[LIndexA].Count);
    end;
  end;

  // initialize entries
  SetLength(Result.Entries, Length(ACollection.ValueSets));
  for LIndexA := 0 to Pred(Length(ACollection.ValueSets)) do
  begin
    Result.Entries[LIndexA].Name := ACollection.ValueSets[LIndexA].Name;

    SetLength(Result.Entries[LIndexA].Values, Length(ACollection.ValueSets[LIndexA].Values));
    for LIndexB := 0 to Pred(Length(ACollection.ValueSets[LIndexA].Values)) do
    begin
      Result.Entries[LIndexA].Values[LIndexB].Name := ACollection.ValueSets[LIndexA].Values[LIndexB];

      SetLength(Result.Entries[LIndexA].Values[LIndexB].Results, Length(ACollection.ResultSet));
      for LIndexC := 0 to Pred(Length(ACollection.ResultSet)) do
      begin
        Result.Entries[LIndexA].Values[LIndexB].Results[LIndexC].Count := 0;
        Result.Entries[LIndexA].Values[LIndexB].Results[LIndexC].Name  := ACollection.ResultSet[LIndexC];

        // fill knowledge base
        for LIndexD := 0 to Pred(Length(ACollection.ValueLists)) do
        begin
          if AnsiLowerCase(ACollection.ValueLists[LIndexD].Result) = AnsiLowerCase(Result.Entries[LIndexA].Values[LIndexB].Results[LIndexC].Name) then
          begin
            for LIndexE := 0 to Pred(Length(ACollection.ValueLists[LIndexD].Values)) do
            begin
              if ((AnsiLowerCase(ACollection.ValueLists[LIndexD].Values[LIndexE].Name) = AnsiLowerCase(Result.Entries[LIndexA].Name)) and
                  (AnsiLowerCase(ACollection.ValueLists[LIndexD].Values[LIndexE].Value) = AnsiLowerCase(Result.Entries[LIndexA].Values[LIndexB].Name))) then
              begin
                Inc(Result.Entries[LIndexA].Values[LIndexB].Results[LIndexC].Count);
                Break;
              end;
            end;
          end;
        end;
      end;
    end;
  end;
end;

function LoadCollection(const AFileName : String; out ACollection : TValueCollection) : Boolean;
  function SeperateList(AList : String; ASeperator : String) : TStringArray;
  begin
    SetLength(Result, 0);

    while (Pos(ASeperator, AList) > 0) do
    begin
      SetLength(Result, Succ(Length(Result)));

      Result[High(Result)] := Trim(Copy(AList, 1, Pred(Pos(ASeperator, AList))));
      if (Length(Result[High(Result)]) <= 0) then
        SetLength(Result, Pred(Length(Result)));

      Delete(AList, 1, Pred(Pos(ASeperator, AList) + Length(ASeperator)));
      AList := Trim(AList);
    end;

    if (Length(AList) > 0) then
    begin
      SetLength(Result, Succ(Length(Result)));

      Result[High(Result)] := Trim(AList);
    end;
  end;
const
  CAssign     = '=';
  CComment    = '//';
  CSetClose   = '}';
  CSetOpen    = '{';
  CSetDivide  = '|';
  CValueClose = ']';
  CValueOpen  = '[';
type
  TLoadStatus = (lsBeforeSet, lsSet, lsBeforeValue, lsValue, lsBeforeEnd);
var
  LFile   : TextFile;
  LFound  : Boolean;
  LIndexA : LongInt;
  LIndexB : LongInt;
  LLine   : String;
  LList   : TStringArray;
  LName   : String;
  LStatus : TLoadStatus;
  LValue  : String;
begin
  Result := false;

  // reset result record
  SetLength(ACollection.ResultSet, 0);
  SetLength(ACollection.ValueSets, 0);
  SetLength(ACollection.ValueLists, 0);

  LStatus := lsBeforeSet;
  if FileExists(AFileName) then
  begin
    AssignFile(LFile, AFileName);

    Reset(LFile);
    try
      while not(EoF(LFile)) do
      begin
        // read next line
        ReadLn(LFile, LLine);

        // delete commented out text
        if (Pos(CComment, LLine) > 0) then
          SetLength(LLine, Pred(Pos(CComment, LLine)));
        LLine := Trim(LLine);

        // ignore empty lines
        if (Length(LLine) > 0) then
        begin
          case LStatus of
            lsBeforeSet :
            begin
              // {..|..}
              if ((Pos(CSetOpen, LLine) = 1) and
                  (Pos(CSetDivide, LLine) > 1) and
                  (Pos(CSetClose, LLine) > Pos(CSetDivide, LLine))) then
              begin
                // delete "{"
                Delete(LLine, 1, 1);

                LList := SeperateList(Copy(LLine, 1, Pred(Pos(CSetClose, LLine))), CSetDivide);
                Delete(LLine, 1, Pos(CSetClose, LLine));
                LLine := Trim(LLine);

                if (Length(LLine) <= 0) then
                begin
                  ACollection.ResultSet := LList;

                  LStatus := lsSet;
                end
                else
                begin
                  // ERROR
                  Result := false;
                  Break;
                end;
              end
              else
              begin
                // ERROR
                Result := false;
                Break;
              end;
            end;

            lsSet :
            begin
              // ..=..
              if (Pos(CAssign, LLine) > 1) then
              begin
                LName := Trim(Copy(LLine, 1, Pred(Pos(CAssign, LLine))));
                Delete(LLine, 1, Pos(CAssign, LLine));
                LValue := Trim(LLine);

                if ((Length(LName) > 0) and (Length(LValue) > 0)) then
                begin
                  // set Name to Value
                  LFound := false;
                  for LIndexA := 0 to Pred(Length(ACollection.ValueSets)) do
                  begin
                    if (AnsiLowerCase(ACollection.ValueSets[LIndexA].Name) = AnsiLowerCase(LName)) then
                    begin
                      for LIndexB := 0 to Pred(Length(ACollection.ValueSets[LIndexA].Values)) do
                      begin
                       LFound := (AnsiLowerCase(ACollection.ValueSets[LIndexA].Values[LIndexB]) = AnsiLowerCase(LValue));
                       if (LFound) then
                         Break;
                      end;

                      if not(LFound) then
                      begin
                        SetLength(ACollection.ValueSets[LIndexA].Values, Succ(Length(ACollection.ValueSets[LIndexA].Values)));
                        ACollection.ValueSets[LIndexA].Values[High(ACollection.ValueSets[LIndexA].Values)] := LValue;

                        LFound := true;
                      end;

                      Break;
                    end;
                  end;

                  if not(LFound) then
                  begin
                    SetLength(ACollection.ValueSets, Succ(Length(ACollection.ValueSets)));
                    ACollection.ValueSets[High(ACollection.ValueSets)].Name := LName;

                    SetLength(ACollection.ValueSets[High(ACollection.ValueSets)].Values, Succ(Length(ACollection.ValueSets[High(ACollection.ValueSets)].Values)));
                    ACollection.ValueSets[High(ACollection.ValueSets)].Values[High(ACollection.ValueSets[High(ACollection.ValueSets)].Values)] := LValue;
                  end;

                  LStatus := lsBeforeValue;
                end
                else
                begin
                  // ERROR
                  Result := false;
                  Break;
                end;
              end
              else
              begin
                // ERROR
                Result := false;
                Break;
              end;
            end;

            lsBeforeValue :
            begin
              // [..]
              if ((Pos(CValueOpen, LLine) = 1) and
                  (Pos(CValueClose, LLine) > 2)) then
              begin
                // delete "["
                Delete(LLine, 1, 1);

                LValue := Trim(Copy(LLine, 1, Pred(Pos(CValueClose, LLine))));
                Delete(LLine, 1, Pos(CValueClose, LLine));
                LLine := Trim(LLine);

                if (Length(LLine) <= 0) then
                begin
                  SetLength(ACollection.ValueLists, Succ(Length(ACollection.ValueLists)));
                  ACollection.ValueLists[High(ACollection.ValueLists)].Result := LValue;

                  LStatus := lsValue;
                end
                else
                begin
                  // ERROR
                  Result := false;
                  Break;
                end;
              end
              else
              begin
                // ..=..
                if (Pos(CAssign, LLine) > 1) then
                begin
                  LName := Trim(Copy(LLine, 1, Pred(Pos(CAssign, LLine))));
                  Delete(LLine, 1, Pos(CAssign, LLine));
                  LValue := Trim(LLine);

                  // set Name to Value
                  LFound := false;
                  for LIndexA := 0 to Pred(Length(ACollection.ValueSets)) do
                  begin
                    if (AnsiLowerCase(ACollection.ValueSets[LIndexA].Name) = AnsiLowerCase(LName)) then
                    begin
                      for LIndexB := 0 to Pred(Length(ACollection.ValueSets[LIndexA].Values)) do
                      begin
                       LFound := (AnsiLowerCase(ACollection.ValueSets[LIndexA].Values[LIndexB]) = AnsiLowerCase(LValue));
                       if (LFound) then
                         Break;
                      end;

                      if not(LFound) then
                      begin
                        SetLength(ACollection.ValueSets[LIndexA].Values, Succ(Length(ACollection.ValueSets[LIndexA].Values)));
                        ACollection.ValueSets[LIndexA].Values[High(ACollection.ValueSets[LIndexA].Values)] := LValue;

                        LFound := true;
                      end;

                      Break;
                    end;
                  end;

                  if not(LFound) then
                  begin
                    SetLength(ACollection.ValueSets, Succ(Length(ACollection.ValueSets)));
                    ACollection.ValueSets[High(ACollection.ValueSets)].Name := LName;

                    SetLength(ACollection.ValueSets[High(ACollection.ValueSets)].Values, Succ(Length(ACollection.ValueSets[High(ACollection.ValueSets)].Values)));
                    ACollection.ValueSets[High(ACollection.ValueSets)].Values[High(ACollection.ValueSets[High(ACollection.ValueSets)].Values)] := LValue;
                  end;
                end
                else
                begin
                  // ERROR
                  Result := false;
                  Break;
                end;
              end;
            end;

            lsValue :
            begin
              // ..=..
              if (Pos(CAssign, LLine) > 1) then
              begin
                LName := Trim(Copy(LLine, 1, Pred(Pos(CAssign, LLine))));
                Delete(LLine, 1, Pos(CAssign, LLine));
                LValue := Trim(LLine);

                // set Name to Value
                LFound := false;
                for LIndexA := 0 to Pred(Length(ACollection.ValueLists[High(ACollection.ValueLists)].Values)) do
                begin
                  if (AnsiLowerCase(ACollection.ValueLists[High(ACollection.ValueLists)].Values[LIndexA].Name) = AnsiLowerCase(LName)) then
                  begin
                    ACollection.ValueLists[High(ACollection.ValueLists)].Values[LIndexA].Value := LValue;

                    LFound := true;
                    Break;
                  end;
                end;

                if not(LFound) then
                begin
                  SetLength(ACollection.ValueLists[High(ACollection.ValueLists)].Values, Succ(Length(ACollection.ValueLists[High(ACollection.ValueLists)].Values)));
                  ACollection.ValueLists[High(ACollection.ValueLists)].Values[High(ACollection.ValueLists[High(ACollection.ValueLists)].Values)].Name  := LName;
                  ACollection.ValueLists[High(ACollection.ValueLists)].Values[High(ACollection.ValueLists[High(ACollection.ValueLists)].Values)].Value := LValue;
                end;

                LStatus := lsBeforeEnd;
                Result := true;
              end
              else
              begin
                // ERROR
                Result := false;
                Break;
              end;
            end;

            lsBeforeEnd :
            begin
              // [..]
              if ((Pos(CValueOpen, LLine) = 1) and
                  (Pos(CValueClose, LLine) > 2)) then
              begin
                // delete "["
                Delete(LLine, 1, 1);

                LValue := Trim(Copy(LLine, 1, Pred(Pos(CValueClose, LLine))));
                Delete(LLine, 1, Pos(CValueClose, LLine));
                LLine := Trim(LLine);

                if (Length(LLine) <= 0) then
                begin
                  SetLength(ACollection.ValueLists, Succ(Length(ACollection.ValueLists)));
                  ACollection.ValueLists[High(ACollection.ValueLists)].Result := LValue;

                  LStatus := lsValue;
                  Result := false;
                end
                else
                begin
                  // ERROR
                  Result := false;
                  Break;
                end;
              end
              else
              begin
                // ..=..
                if (Pos(CAssign, LLine) > 1) then
                begin
                  LName := Trim(Copy(LLine, 1, Pred(Pos(CAssign, LLine))));
                  Delete(LLine, 1, Pos(CAssign, LLine));
                  LValue := Trim(LLine);

                  // set Name to Value
                  LFound := false;
                  for LIndexA := 0 to Pred(Length(ACollection.ValueLists[High(ACollection.ValueLists)].Values)) do
                  begin
                    if (AnsiLowerCase(ACollection.ValueLists[High(ACollection.ValueLists)].Values[LIndexA].Name) = AnsiLowerCase(LName)) then
                    begin
                      ACollection.ValueLists[High(ACollection.ValueLists)].Values[LIndexA].Value := LValue;

                      LFound := true;
                      Break;
                    end;
                  end;

                  if not(LFound) then
                  begin
                    SetLength(ACollection.ValueLists[High(ACollection.ValueLists)].Values, Succ(Length(ACollection.ValueLists[High(ACollection.ValueLists)].Values)));
                    ACollection.ValueLists[High(ACollection.ValueLists)].Values[High(ACollection.ValueLists[High(ACollection.ValueLists)].Values)].Name  := LName;
                    ACollection.ValueLists[High(ACollection.ValueLists)].Values[High(ACollection.ValueLists[High(ACollection.ValueLists)].Values)].Value := LValue;
                  end;
                end
                else
                begin
                  // ERROR
                  Result := false;
                  Break;
                end;
              end;
            end;
          end;
        end;
      end;
    finally
      Close(LFile);
    end;
  end;
end;

function SelectCaseResults(const ABase : TKnowledgeBase; const ACase : TValueAssigns; out AResults : TResultLists) : Boolean;
var
  LIndexA : LongInt;
  LIndexB : LongInt;
  LIndexC : LongInt;
  LIndexD : LongInt;
begin
  Result := false;

  for LIndexA := 0 to Pred(Length(ABase.Entries)) do
  begin
    Result := false;
    for LIndexB := 0 to Pred(Length(ACase)) do
    begin
      if (AnsiLowerCase(ACase[LIndexB].Name) = AnsiLowerCase(ABase.Entries[LIndexA].Name)) then
      begin
        for LIndexC := 0 to Pred(Length(ABase.Entries[LIndexA].Values)) do
        begin
          if (AnsiLowerCase(ACase[LIndexB].Value) = AnsiLowerCase(ABase.Entries[LIndexA].Values[LIndexC].Name)) then
          begin
            // initialize result
            if (Length(AResults) <> Length(ABase.Entries[LIndexA].Values[LIndexC].Results)) then
            begin
              SetLength(AResults, Length(ABase.Entries[LIndexA].Values[LIndexC].Results));

              for LIndexD := 0 to Pred(Length(ABase.Entries[LIndexA].Values[LIndexC].Results)) do
              begin
                SetLength(AResults[LIndexD].Values, Length(ABase.Entries));

                AResults[LIndexD].Count  := ABase.Results[LIndexD].Count;
                AResults[LIndexD].Result := ABase.Entries[LIndexA].Values[LIndexC].Results[LIndexD].Name;
              end;
            end;

            // set result
            for LIndexD := 0 to Pred(Length(AResults)) do
            begin
              AResults[LIndexD].Values[LIndexA].Count := ABase.Entries[LIndexA].Values[LIndexC].Results[LIndexD].Count;
              AResults[LIndexD].Values[LIndexA].Name  := ABase.Entries[LIndexA].Name;
            end;

            Result := true;
          end;
        end;
      end;
    end;

    if not(Result) then
    begin
      // ERROR
      Break;
    end;
  end;
end;

end.
